@@ -20,6 +20,7 @@ from . elements import (
     Integer,
     List,
     Location,
+    Nilsimsa,
     NumericRange,
     Schema,
     String,
@@ -53,7 +54,8 @@ def open(path_or_db,
          writable=True,
          overwrite=False,
          spelling=True,
-         replicated=False):
+         replicated=False,
+         inmem=False):
     """Return an xodb database with the given path or xapian database object.
 
     :param path_or_db: A path to a database file or a pre-existing
@@ -71,7 +73,8 @@ def open(path_or_db,
                     writable=writable,
                     overwrite=overwrite,
                     spelling=spelling,
-                    replicated=replicated)
+                    replicated=replicated,
+                    inmem=inmem)
 
 
 def temp(spelling=True):
@@ -87,4 +90,4 @@ def inmemory():
     """Returns an xodb database backed by an in-memory xapian
     database.  Does not support spelling correction.
     """
-    return open(xapian.inmemory_open(), spelling=False)
+    return open(xapian.inmemory_open(), spelling=False, inmem=True)
