@@ -1,22 +1,23 @@
-from __future__ import absolute_import
-
 import time
 import string
 import logging
+from uuid import uuid4
 from collections import namedtuple
 from contextlib import contextmanager
 
 import xapian
+import nilsimsa
 
 from operator import itemgetter
 from functools import partial
 from json import loads
 
-from xapian import Query, QueryParser, DocNotFoundError
+from xapian import QueryParser, DocNotFoundError
 
 from . import snowball
+from .elements import Schema, String, Integer, Float, Array
 from .exc import ValidationError, PrefixError
-from .tools import LRUDict
+from .tools import LRUDict, lazy_property
 
 
 RETRY_LIMIT = 5
@@ -62,15 +63,67 @@ def to_term(value, prefix=None):
     return _prefix(prefix) + value if prefix else value
 
 
+class QuerySchema(Schema):
+    type = String.using(default='xodbquery')
+    term = Array.of(String).using(prefix='query_term',
+                                  getter=lambda s, o, e: list(o))
+    id = Array.of(String).using(prefix='query_id',
+                                getter=lambda s, o, e: o.__xodb_id__)
+
+
+class Query(object):
+    __xodb_schema__ = QuerySchema
+
+    class __metaclass__(type):
+
+        def __getattr__(cls, name):
+            return getattr(xapian.Query, name)
+
+    def __init__(self, *args, **kwargs):
+        self.__xodb_query__ = kwargs.pop('query', None) or xapian.Query(*args, **kwargs)
+        self.__xodb_id__ = uuid4()
+
+    def __getattr__(self, name):
+        return getattr(self.__xodb_query__, name)
+
+    def __getstate__(self):
+        return dict(query=self.__xodb_query__.serialise(),
+                    id=str(self.__xodb_id__))
+
+    def __setstate__(self, state):
+        self.__xodb_query__ = xapian_Query.unserialise(state['query'])
+        self.__xodb_id__ = uuid.UUID(state['id'])
+
+    def __str__(self):
+        return str(self.__xodb_query__)
+
+    def __iter__(self):
+        return iter(self.__xodb_query__)
+
+
+class RecordSchema(Schema):
+    type = String.using(default='xodbrecord')
+    rank = Integer.using(getter=lambda s, o, e: o._xodb_rank)
+    percent = Integer.using(getter=lambda s, o, e: o._xodb_percent)
+    weight = Integer.using(getter=lambda s, o, e: o._xodb_weight)
+
+
 class Record(object):
     """Nice attribute-accessable record for a search result."""
+    __xodb_schema__ = RecordSchema
 
-    def __init__(self, schema, document, percent, rank, weight):
-        self._xodb_schema = schema
+    def __init__(self, document, percent, rank, weight, query, db):
         self._xodb_document = document
         self._xodb_percent = percent
         self._xodb_rank = rank
         self._xodb_weight = weight
+        self._xodb_query = query
+        self._xodb_db = db
+
+    @lazy_property
+    def _xodb_schema(self):
+        typ, data = loads(self._xodb_document.get_data())
+        return _lookup_schema(typ).from_flat(data)
 
     def __getattr__(self, name):
         try:
@@ -81,9 +134,20 @@ class Record(object):
     def __repr__(self):
         return repr(self._xodb_schema)
 
+    def __getstate__(self):
+        state = self.__dict__.copy()
+        if '_xodb_schema' in state:
+            del state['_xodb_schema']
+        state['_xodb_document'] = self._xodb_document.serialise()
+        return state
 
-def record_factory(database, data, doc, percent, rank, weight):
-    return Record(data, doc, percent, rank, weight)
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+        self._xodb_document = xapian.Document.unserialise(self._xodb_document)
+
+
+def record_factory(database, doc, percent, rank, weight, query, db):
+    return Record(doc, percent, rank, weight, query, db)
 
 
 class LanguageDecider(xapian.ExpandDecider):
@@ -293,6 +357,8 @@ class Database(object):
     def schema_for(self, otype):
         """Get the schema for a given type, or one of its
         superclasses."""
+        if hasattr(otype, '__xodb_schema__'):
+            return otype.__xodb_schema__
         for base in otype.__mro__:
             if base in self.type_map:
                 return self.type_map[base]
@@ -490,6 +556,7 @@ class Database(object):
             else:
                 schema_type = self.schema_for(type(obj))
         schema = schema_type.from_defaults()
+        schema.__xodb_db__ = self
         schema.update_by_object(obj)
 
         if validate and not schema.validate():
@@ -508,8 +575,9 @@ class Database(object):
         Convienient wrapper that does the object->schema->document
         transformation.
         """
-        schema = self.to_schema(obj, validate, schema_type)
-        return self.doc_from_dict(schema.memo.dict)
+        if not isinstance(obj, Schema):
+            obj = self.to_schema(obj, validate, schema_type)
+        return self.doc_from_dict(obj.__xodb_memo__.dict)
 
     def doc_from_dict(self, data):
         """Take an intermediate representation of a document (a
@@ -528,8 +596,6 @@ class Database(object):
                 term, typ = term_tup
             if ':' in term:
                 prefix, _, value = term.partition(':')
-                # FIXME: check_prefix has a second param called boolean=False
-                # Is passing the typ in the right thing to do?
                 self.check_prefix(prefix, typ)
                 term = 'X%s:%s' % (prefix.upper(), value)
             doc.add_term(term, wdfinc)
@@ -767,7 +833,7 @@ class Database(object):
                 qp = self.get_query_parser(language, default_op, 
                                            retry_limit=retry_limit)
                 def query_op():
-                    return qp.parse_query(query, parser_flags)
+                    return Query(query=qp.parse_query(query, parser_flags))
                 result = self.retry_if_modified(query_op, retry_limit)
                 if not self.inmem:
                     self.query_cache[cache_key] = result
@@ -792,6 +858,9 @@ class Database(object):
               match_spy=None,
               document=False,
               echo=False,
+              disimilate=False,
+              disimilate_field='nilsimsa',
+              disimilate_threshold=100,
               parser_flags=default_parser_flags,
               default_op=Query.OP_AND,
               retry_limit=RETRY_LIMIT):
@@ -809,7 +878,7 @@ class Database(object):
                              default_op, parser_flags)
         if echo:
             print "Done parsing query: %s" % str(query)
-        enq.set_query(query)
+        enq.set_query(query.__xodb_query__)
 
         limit = limit or self.backend.get_doccount()
 
@@ -823,6 +892,8 @@ class Database(object):
 
         tries = 0
         seen = set()
+        disimilator = set()
+        sim_comp = nilsimsa.compare_hexdigests
         while True:
             try:
                 # _build_mset may retry internally on DatabaseError
@@ -851,12 +922,25 @@ class Database(object):
                             continue
                         typ, data = loads(data)
                         seen.add(docid)
-                        yield self.record_factory(_lookup_schema(typ).from_flat(data),
-                                                  doc,
-                                                  record.percent,
-                                                  record.rank,
-                                                  record.weight,
-                                                  )
+                        record = self.record_factory(doc,
+                                                     record.percent,
+                                                     record.rank,
+                                                     record.weight,
+                                                     query,
+                                                     self,
+                                                     )
+                        if disimilate:
+                            yield_it = True
+                            rhash = getattr(record, disimilate_field, None)
+                            if rhash:
+                                if any((sim_comp(rhash, h) > disimilate_threshold)
+                                       for h in set(disimilator)):
+                                    yield_it = False
+                            if yield_it:
+                                disimilator.add(rhash)
+                                yield record
+                        else:
+                            yield record
                 # no errors exhuasting the set? break out and we're done
                 break
             except xapian.DatabaseError, e:
@@ -890,7 +974,7 @@ class Database(object):
         if echo:
             print str(query)
         enq = xapian.Enquire(self.backend)
-        enq.set_query(query)
+        enq.set_query(query.__xodb_query__)
 
         mset = self._build_mset(enq, retry_limit=retry_limit)
         return mset.size()
@@ -1007,7 +1091,7 @@ class Database(object):
                              default_op, parser_flags,
                              retry_limit=retry_limit)
 
-        enq.set_query(query)
+        enq.set_query(query.__xodb_query__)
         op = lambda: enq.get_mset(0, 0, limit)
         mset = self.retry_if_modified(op, retry_limit)
 
@@ -1065,7 +1149,8 @@ class Database(object):
                 default_op=Query.OP_AND,
                 parser_flags=default_parser_flags,
                 retry_limit=RETRY_LIMIT,
-                format_term=True):
+                format_term=True,
+                collapse_stems=True):
         """
         Suggest terms that would possibly yield more relevant results
         for the given query.
@@ -1080,7 +1165,7 @@ class Database(object):
 
         if echo:
             print str(query)
-        enq.set_query(query)
+        enq.set_query(query.__xodb_query__)
 
         mset = self._build_mset(enq, offset=moffset, limit=mlimit,
                                 retry_limit=retry_limit)
@@ -1109,6 +1194,12 @@ class Database(object):
                 enq.INCLUDE_QUERY_TERMS,
                 1.0, decider, -3)
 
+        stemmer = None
+        stems = None
+        if collapse_stems:
+            stems = set()
+            stemmer = xapian.Stem(language)
+
         eset = self.retry_if_modified(op, retry_limit)
 
         for item in eset.items:
@@ -1119,6 +1210,10 @@ class Database(object):
                     val = '%s:"%s"' % (prefix, suffix)
                 else:
                     val = '%s:%s' % (prefix, suffix)
+            if collapse_stems:
+                if stemmer(val) in stems:
+                    continue
+                stems.add(stemmer(val))
             if score:
                 yield (val, item[1])
             else:
