@@ -2,7 +2,9 @@ import logging
 import unicodedata
 import translitcodec
 
+import nilsimsa
 from flatland import schema
+from flatland.schema.forms import _MetaForm
 from json import dumps
 
 from . memo import Memo
@@ -32,7 +34,13 @@ def _normalize(value):
 _use_schema = object() # marker says use schema term generator
 
 
-class Schema(schema.Form):
+class SparseForm(schema.SparseDict):
+    __metaclass__ = _MetaForm
+
+    minimum_fields = 'required'
+
+
+class Schema(SparseForm):
 
     language = None
     """The overarching language for this schema.
@@ -55,6 +63,8 @@ class Schema(schema.Form):
     When False (default) an invalid term in a document raises InvalidTermError.
     """
 
+    __xodb_db__ = None
+
     def update_by_object(self, obj):
         """Updates fields with an object's attributes.
 
@@ -64,14 +74,23 @@ class Schema(schema.Form):
         *obj*.  Object attributes that do not correspond to field names are
         ignored.
         """
-        for name, element in self.iteritems():
+        for schema in self.field_schema:
+            name = schema.name
+            element = schema.from_defaults()
+            self[name] = element.value
+            value = None
             if element.getter is not None:
-                self[name] = element.getter(self, obj, element)
+                value = element.getter(self, obj, element)
             elif hasattr(obj, name):
-                self[name] = getattr(obj, name)
+                value = getattr(obj, name)
+            if value is None and not element.is_empty:
+                if element.optional:
+                    del self[name]
+            else:
+                self[name] = value
 
     @property
-    def memo(self):
+    def __xodb_memo__(self):
         """
         Walk the children, indexing each one with a handler into a
         memo object.
@@ -85,12 +104,12 @@ class Schema(schema.Form):
 
         self._memo = Memo()
         self._memo.set_lang(lang)
-        self._handle_children(self)
+        self._handle_children(self, None)
         self._memo.data = dumps((_schema_name(self), self.flatten()))
         return self._memo
 
-    def _handle_children(self, parent):
-        for el in parent.all_children:
+    def _handle_children(self, parent, grandparent):
+        for el in parent.children:
             if el.index:
                 if isinstance(el, String):
                     handler = self._handle_string
@@ -110,6 +129,8 @@ class Schema(schema.Form):
                     handler = self._handle_numericrange
                 elif isinstance(el, Location):
                     handler = self._handle_location
+                elif isinstance(el, Nilsimsa):
+                    handler = self._handle_nilsimsa
                 elif isinstance(el, List):
                     handler = self._handle_children
                 elif isinstance(el, Dict):
@@ -121,7 +142,7 @@ class Schema(schema.Form):
 
                 value = None
                 try:
-                    value = handler(el)
+                    value = handler(el, parent)
                 except InvalidTermError, e:
                     if self.ignore_invalid_terms:
                         logger.warning('Invalid term ignored: %r' % el)
@@ -135,6 +156,8 @@ class Schema(schema.Form):
                             self._memo.add_term(
                                 _prefix(self.facet_prefix, term),
                                 True)
+            if not el.store:
+                el.value = None
         return True
 
     def _handle_scalar(self, term, value, element, type=None):
@@ -151,40 +174,40 @@ class Schema(schema.Form):
                 memo.add_value(name, value, type)
             return value
 
-    def _handle_string(self, element):
+    def _handle_string(self, element, parent):
         term = element.u
         value = element.value
         if value:
             return self._handle_scalar(term, value, element, 'string')
 
-    def _handle_integer(self, element):
+    def _handle_integer(self, element, parent):
         term = element.u
         value = element.value
         if value:
             return self._handle_scalar(term, value, element, 'integer')
 
-    def _handle_float(self, element):
+    def _handle_float(self, element, parent):
         # TODO:mp floats are currently storage only
         pass
 
-    def _handle_boolean(self, element):
+    def _handle_boolean(self, element, parent):
         value = 'true' if element.value else 'false'
         if value:
             return self._handle_scalar(value, value, element, 'integer')
 
-    def _handle_date(self, element):
+    def _handle_date(self, element, parent):
         if element.value:
             term = element.value.strftime(element.term_format)
             value = element.value.strftime(element.value_format)
             return self._handle_scalar(term, value, element, 'date')
 
-    def _handle_datetime(self, element):
+    def _handle_datetime(self, element, parent):
         if element.value:
             term = element.value.strftime(element.term_format)
             value = element.value.strftime(element.value_format)
             return self._handle_scalar(term, value, element, 'datetime')
 
-    def _handle_text(self, element):
+    def _handle_text(self, element, parent):
         value = element.value
         if value is None:
             return
@@ -225,7 +248,7 @@ class Schema(schema.Form):
                       element.position_start)
         return value
 
-    def _handle_numericrange(self, element):
+    def _handle_numericrange(self, element, parent):
         maxv = element['high'].value or 0
         minv = element['low'].value or 0
         if minv == maxv == 0:
@@ -247,7 +270,7 @@ class Schema(schema.Form):
                                 element.wdf_inc)
         return True
 
-    def _handle_location(self, element):
+    def _handle_location(self, element, parent):
         memo = self._memo
         h = 'loc_' + element.hash(element.radians)
         memo.add_term(h, element.boolean, element.wdf_inc)
@@ -255,6 +278,20 @@ class Schema(schema.Form):
             memo.add_value(element.name, h, 'location')
         return True
 
+    def _handle_nilsimsa(self, element, parent):
+        memo = self._memo
+        try:
+            val = (parent[element.from_field].value
+                   .encode('translit/long')
+                   .encode('ascii', 'ignore'))
+            element.u = nilsimsa.Nilsimsa([val]).hexdigest()
+        except Exception:
+            print 'whoops: ', val
+            return
+        if element.u:
+            return self._handle_scalar(element.u, element.u, element, 'string')
+        return True
+
 
 class _BaseElement(object):
 
@@ -265,6 +302,12 @@ class _BaseElement(object):
     do not generate terms or values.
     """
 
+    store = True
+    """If True, store this elements value when the schema is saved.
+
+    If False, this element is not stored on the schema.
+    """
+
     getter = None
     """Getter function to calculate value from a given object.
 
@@ -480,6 +523,15 @@ class NumericRange(_BaseRange):
     """
 
 
+class Nilsimsa(schema.String, _BaseElement):
+
+    from_field = None
+    """What other field the simhash should be computed from.
+
+    If this is not provided, indexing will raise an error.
+    """
+
+
 class Location(schema.Compound, _BaseElement):
     """ Compound location is a 2-tuple of lat/lon coordinates.
     """
